{                BEST ATTAINABLE MATCH ALGORITHM
                          R.M.Fewster
              University of St. Andrews, Scotland, UK.

                             bam.p
                             -----
Pascal program to compute the Best Attainable Match for a pair of binary
(presence/absence) spatial maps.  This program is free and comes with
no guarantees.

Edit the program at the 3 points marked *EDIT* to give upper limit for
the number of mismatched 0's, upper limit for the number of mismatched
1's, and total number of sites.  In most cases (or if in doubt) all three
numbers will be equal to the number of sites in the survey.  For greater
economy with memory usage, replace the first two numbers with the first
two entries of Nmismfile, obtained as output from the program bamprelim.p.  

After the code is edited, compile and run the program bamprelim.p, then
compile and run this program.  Recompiling on subsequent runs is only
necessary if the code has been changed.} 


program bestattainablematch(Nmismfile, mismatch0, mismatch1, opmismatch);

const Nmism0lim=2500; {*EDIT* : upper limit for number of mismatched 0's}
      Nmism1lim=2500; {*EDIT* : upper limit for number of mismatched 1's}
      Ntotal=2500;    {*EDIT* : total number of sites in each map}

type
   siterange = 1..Ntotal;
   cliquesets = set of siterange;
   cliquearray = array[siterange] of cliquesets;
   header0 = array[1..Nmism0lim] of char;
   header1 = array[1..Nmism1lim] of char;


var opmism : cliquearray;
          {opmism holds the mismatch numbers for opposite mismatches in 
          cliques for each mismatched site}

    Nmism0, Nmism1, N : integer;   {the actual numbers of mismatched 0's
                                    and 1's to be read in from a file}   
    mism0, mism1, cc, block : cliquesets;
       {mism0 and mism1 hold mismatch numbers of mismatched 0's and 1's resp.}
                                     {cc is the current connected clique}
                                      {block is the current block}
    valency, Bnumber : array[1..Ntotal] of integer;  
                                 {valency holds number of opposite mismatches}
         {Bnumber[i]=b if i is the b'th mismatched 0 or the b'th mismatched 1}

    clus0 : array[1..Nmism0lim] of integer; 
    clus1 : array[1..Nmism1lim] of integer;
                    {clus0[i]=hh if the site with mismatch number hh is the
                     i'th zero in the current connected clique (cluster)}

    i,j,k,t, cst :integer;

    mismatch0, mismatch1, opmismatch, Nmismfile: text;  
                                         {info output from bamprelim.p}
    c, h : integer;

    head0 : header0;
    head1 : header1;
    BAM, ninblock, r : integer;   {used in BAM algorithm}
    timetostop, nochains, foundblock, singagain : boolean;
    Nsing1, Nsing, nones, nzeros, nblocks, n2blocks, clusswap:integer;
                                                 {counters for summary}


{-----PROCEDURES------------------------------------------------------}

procedure singletons;  {only operates on the whole matrix B and the full 
                        valency vector: swaps all sites of valency 1}
var someleft : boolean;
    l, ndone : integer;

  {----swap procedure-----}

  procedure swap(sti:integer); {to swap the singleton, sti}
  var stk, m : integer;
  begin  {swap}
    Nsing:=Nsing+1; ndone:=ndone+1;
    stk:=0;
    repeat
      stk:=stk+1;
    until stk in opmism[sti];  {found stk to swap with sti}
    
    {if valency[stk]=>1 then must adjust the valencies and opmism sets of 
    the other sites valent with stk:}

    if valency[stk]>1 then for m:=1 to N do 
       if (m in opmism[stk]) and (m<>sti) then begin
          opmism[m]:=opmism[m]-[stk];
          valency[m]:=valency[m]-1;
          if (valency[m]=1) and (m<sti) then someleft:=true;
              {swapping stk has left other singletons behind sti}
       end;  {if}

    {adjust valencies and headers of sti and stk}
    valency[sti]:=-1; valency[stk]:=-1;
    if sti in mism0 then begin  {sti is a mismatched 0}
       head0[Bnumber[sti]]:=' '; head1[Bnumber[stk]]:=' ';
    end
    else begin    {sti is a mismatched 1}
       head1[Bnumber[sti]]:=' '; head0[Bnumber[stk]]:=' ';
    end;   {if}
    opmism[stk]:=opmism[stk]-[sti];

  end;  {swap}  

{---singletons---------}

begin  {singletons}
  someleft:=true; ndone:=0;
  while someleft=true do begin
     l:=0; someleft:=false;
     repeat
        l:=l+1;
        if valency[l]=1 then swap(l);
     until l=N;
  end;  {while}
 {***} writeln('Singletons: swap=', ndone:2); 

end;  {singletons}
   

{------------------------------------}

procedure TidyB;  

{all rows/columns of B with valency 0 will be given a 
blank header, have valency changed to -1, and BAM reduced by 1. 
Procedure then checks whether there are any more sites left to be swapped.} 
 
var l:integer;

begin  {TidyB}
{***} { write('TidyB....'); }
   for l:=1 to N do if valency[l]=0 then begin
      BAM:=BAM-1;
       valency[l]:=-1;
       if l in mism0 then head0[Bnumber[l]]:=' ' else head1[Bnumber[l]]:=' ';
   end;   {l}
   timetostop:=true;
   l:=0;
   if Nmism0 > 0 then repeat
      l:=l+1;
       if head0[l]='u' then timetostop:=false;
   until (head0[l]='u') or (l=Nmism0);
               {note: if there are no 'u's among head0[l] then there 
               can be none among head1[l] since these sites would then
               have valency 0 and would have been removed in previous step.}
{***}{writeln('Done'); }
end;   {TidyB}


{------------------------------------}

procedure find_connected_clique(site:integer);

var mark : array[1..Ntotal] of integer;
    hh : integer;
    foundclique : boolean;

begin
    for hh:=1 to N do mark[hh]:=0;
    cc:=[site];
    repeat
       foundclique:=true;
       for hh:=1 to N do if (hh in cc) and (mark[hh]=0) then begin
           mark[hh]:=1;
           if not (opmism[hh]<=cc) then begin
              cc:=cc+opmism[hh];
              foundclique:=false;  {the status of foundclique is only 
                                   changed if new elements have been 
                                                         added to cc}
           end; {if}
       end;  {hh}
    until foundclique=true;
    {the connected clique containing site is now found}

    {change the headers accordingly}
    nzeros:=0; nones:=0;   
    for hh:=1 to N do if hh in cc then begin
        if hh in mism0 then begin
             head0[Bnumber[hh]]:='c';
             nzeros:=nzeros+1;
             clus0[nzeros]:=hh;
        end 
        else begin
             head1[Bnumber[hh]]:='c';
             nones:=nones+1;
             clus1[nones]:=hh;
        end;
    end;  {hh}
    if nzeros<Nmism0 then for hh:=nzeros+1 to Nmism0 do clus0[hh]:=0;
    if nones<Nmism1 then for hh:=nones+1 to Nmism1 do clus1[hh]:=0;

end;  {find_connected_clique}    


{----------------------------------}

procedure check_chains;  

{determines whether the current connected clique forms a 'chain', ie. 
whether all sites have the same valency. If so, swaps clique}

var hh, val : integer;

begin
    val:=valency[clus0[1]];    {established the valency of a site in cc, val}

    nochains:=false; hh:=1;
    repeat
       hh:=hh+1;
       if valency[clus0[hh]]<>val then nochains:=true;
    until (nochains=true) or (hh=nzeros);  {now know whether all 0's in cc 
                                            have same valency}  
    if nochains=false then begin  {now check the 1's in cc}
       hh:=0;
       repeat
           hh:=hh+1;
           if valency[clus1[hh]]<>val then nochains:=true;
       until (nochains=true) or (hh=nones);  {checked all 1's in cc}
    end;   {if}   
      {all sites in cc checked: know whether or not a chain exists (ie cc 
      is perfect) }

    if nochains=false then begin   {swap the perfect cluster cc}

       {set all headers of sites in cc to blanks, and all valencies to -1}
       {BAM is unaffected because we obtain full swap}

       for hh:=1 to nzeros do begin
           head0[Bnumber[clus0[hh]]]:=' '; 
           valency[clus0[hh]]:=-1;
       end;   {hh a zero in cc}
       for hh:=1 to nones do begin
           head1[Bnumber[clus1[hh]]]:=' ';
           valency[clus1[hh]]:=-1;
       end;   {hh a one in cc}
      clusswap:=clusswap+nzeros;  {swap attributable to perfect clusters
                                   increases by # 0's (= # 1's) in cluster}
{***} writeln('Swapped exact connected component: valency=', val:1,', swap=',nones:2); 

    end;  {if cc perfect}
end;  {check_chains} 


{---------------------------------}

procedure RestoreB;  {simply replaces any remaining 'c' headers by 'u's
                     prior to performing the singletons procedure}
var hh : integer;
begin
{***}{write('RestoreB.......');}
     for hh:=1 to nzeros do if head0[Bnumber[clus0[hh]]]='c' 
     then head0[Bnumber[clus0[hh]]]:='u';
     for hh:=1 to nones do if head1[Bnumber[clus1[hh]]]='c'
     then head1[Bnumber[clus1[hh]]]:='u';
{***}{writeln('Done');}
end;  {RestoreB}




{---------------------------------------------------}


procedure find_columnblocks(Lhead0: header0;  Lhead1: header1; 
                            t, s, permission: integer);
                       
var nrows, ncols, r0, ii, tv: integer;  
                                  {r0 is a row with maximum valency in Bm}
                 Chead0 : header0;
                 Chead1 : header1;
                tempval : array[1..Ntotal] of integer;
         
{procedure looks for (t,s)-blocks in the COLUMNS of B, i.e. a set of >=s
rows with entries exclusively within some subset of t columns}


    { *** note that the (t,s)-block here uses different terminology from
     that used in the write-up: there, an (r,s) block had r sites of one
     sort and s>=r of the other; here, a (t,s)-block has t sites of one
     sort and >=s (we don't stipulate how many, just that s is the minimum)
     of the other ***} 


{-----------------------}
  procedure find_tempval;  {finds all valencies for current Bm} 
  var v,w : integer;

  begin  {find_tempval}
     ncols:=0; nrows:=0;
{***}{writeln('tempval...'); }

     for v:=1 to N do tempval[v]:=-1;
     for v:=1 to nzeros do tempval[clus0[v]]:=0;
     for v:=1 to nones do tempval[clus1[v]]:=0; 

     for v:=1 to nzeros do begin
       if (Lhead0[Bnumber[clus0[v]]]='c') then begin
          ncols:=ncols+1;
          for w:=1 to nones do 
             if (Lhead1[Bnumber[clus1[w]]]='c') and 
             (clus1[w] in opmism[clus0[v]])
             then tempval[clus0[v]]:=tempval[clus0[v]]+1;
       end {header 'c'}
       else tempval[clus0[v]]:=-1;  {header not 'c'}
     end;  {v in 1 to nzeros}
     
     for v:=1 to nones do begin        
       if (Lhead1[Bnumber[clus1[v]]]='c') then begin
          nrows:=nrows+1;
          for w:=1 to nzeros do
             if (Lhead0[Bnumber[clus0[w]]]='c') and 
            (clus0[w] in opmism[clus1[v]])
             then tempval[clus1[v]]:=tempval[clus1[v]]+1;
       end {header 'c'}
       else tempval[clus1[v]]:=-1;  {header not 'c'}
     end;  {v in 1 to nones}

    {writeln('found tempval array'); 
    {tempval array found, and ncols, nrows calculated}
 
  end;   {find_tempval}


{-------------------------}
  procedure delete_rows_with_tempval_too_big;

  {rows with tempval>t cannot form part of a (t,s)-block, so are marked 'd'}

  var v,w : integer;
 
  begin  {delete_rows_with_tempval_too_big}
{***}{writeln('rows with tempval too big....');}
  v:=0;
  repeat
     v:=v+1;
     if (Lhead1[Bnumber[clus1[v]]]='c') and (tempval[clus1[v]]>t) then begin
        Lhead1[Bnumber[clus1[v]]]:='d';
        tempval[clus1[v]]:=-1;
        nrows:=nrows-1;
        for w:=1 to nzeros do if (Lhead0[Bnumber[clus0[w]]]='c') 
        and (clus0[w] in opmism[clus1[v]]) then 
        tempval[clus0[w]]:=tempval[clus0[w]]-1;
     end;  {if clus1[v] has header 'c' and tempval>t}
  until (v=nones) or (nrows<s);
 {writeln('deleted rows with tempval too big');}
  end;  {delete_rows_with_tempval_too_big}


{-------------------------}
  procedure delete_columns_with_one_entry;
  
 {if permission=1 then we delete all columns with only one entry, and 
 the associated row, as these cannot form part of a block with no smaller 
 subblocks}

  var l : integer;
      someleft : boolean;  

  procedure delete(col : integer);   {col is the MISMATCH number of the site
                                     corresponding to a column with one entry}
    var row, m, done : integer;
    begin               {delete sub-procedure}
       Lhead0[Bnumber[clus0[col]]]:='d'; tempval[clus0[col]]:=-1; 
       ncols:=ncols-1;
       row:=0; done:=0;
       repeat {find row intersecting with col}
          row:=row+1;
          if (Lhead1[Bnumber[clus1[row]]]='c') and 
          (clus1[row] in opmism[clus0[col]]) then done:=1;
       until done=1;

       {delete the intersecting row}
       Lhead1[Bnumber[clus1[row]]]:='d'; tempval[clus1[row]]:=-1; 
       nrows:=nrows-1;

       for m:=1 to nzeros do if (Lhead0[Bnumber[clus0[m]]]='c') 
       and (clus0[m] in opmism[clus1[row]]) then begin
          tempval[clus0[m]]:=tempval[clus0[m]]-1;
          if (m<col) and (tempval[clus0[m]]=1) then someleft:=true;
                      {m<col iff Bnumber[clus0[m]]<Bnumber[clus0[col]]}
       end;  {m}
    end;  {delete}

    
  begin  {delete_columns_with_one_entry}
{***}{writeln('columns with 1 entry...');}
    repeat
       l:=0; someleft:=false;
       repeat
          l:=l+1;
          if (Lhead0[Bnumber[clus0[l]]]='c') and (tempval[clus0[l]]=1) 
          then delete(l);    
          if (ncols<t) or (nrows<s) then foundblock:=false;
       until (l=nzeros) or (foundblock=false);
    until (someleft=false) or (foundblock=false);
   {writeln('deleted columns with one entry'); }
  end;  {delete_columns_with_one_entry}
  
  
{------------------------}
  procedure delete_columns_with_no_entries; 
  var v : integer;
  begin    {delete_columns_with_no_entries}
{***}{writeln('columns with no entries....');}
    for v:=1 to nzeros do 
       if (Lhead0[Bnumber[clus0[v]]]='c') and (tempval[clus0[v]]=0) 
       then begin
          Lhead0[Bnumber[clus0[v]]]:='d'; tempval[clus0[v]]:=-1; 
          ncols:=ncols-1;
       end;
   {writeln('deleted columns with no entries');}
  end;     {delete_columns_with_no_entries}


{--------------------}
  procedure find_row;
  {finds row r0 with maximum value of tempval in Bm and adds it together
   with its associated columns to the set block}

   var max,v : integer;
   
   begin  {find_row}
    { write('find_row');}
     max:=0;
     for v:=1 to nones do if (Lhead1[Bnumber[clus1[v]]]='c') 
     and (tempval[clus1[v]]>max) then begin
         r0:=clus1[v]; max:=tempval[clus1[v]];  
                                   {r0 is the MISMATCH number of the row}
     end;  {v}
     {found r0 with maximum tempval amongst rows with header 'c'}
     
     Chead0:=Lhead0; Chead1:=Lhead1;
     Chead1[Bnumber[r0]]:='d';   {delete r0 in new matrix C}
    
     block:=block+[r0]; ninblock:=ninblock+1;
     tv:=max; tempval[r0]:=-1;    {** I mean this: r0, not clus1[r0] **}
     for v:=1 to nzeros do if (Lhead0[Bnumber[clus0[v]]]='c') 
     and (clus0[v] in opmism[r0]) then begin
         block:=block+[clus0[v]];  {** I mean this too: clus0[v], not v **}
         ninblock:=ninblock+1;
         Chead0[Bnumber[clus0[v]]]:='d'; tempval[clus0[v]]:=-1; 
                                 {delete columns intersecting r0 in C}
     end;   {v}
     {block now contains r0 and its intersected columns}
     {writeln('found row: r0 = ',r0:2);}
{***} {writeln('done');}
   end;    {find_row}

  
{-----------------------------}
  
begin  {main body of find_columnblocks(Lhead0, Lhead1, t, s, permission) }
  { if permission=1 then writeln('looking for ',t:2,'-blocks') 
   else writeln('looking for ',t:2,'-subblocks'); }
   foundblock:=true;
   find_tempval;
   delete_rows_with_tempval_too_big;
                          {now have nrows rows and ncols columns left}

   if t=0 then begin    {degenerate case :t=0}
      if nrows>=s then begin
         foundblock:=true;
         for ii:=1 to nones do if (Lhead1[Bnumber[clus1[ii]]]='c') 
         then begin
             block:=block+[clus1[ii]];
             ninblock:=ninblock+1;
         end; {ii}
      end  {nrows>=s}
      else foundblock:=false;
   end   {t=0}
   else begin  {general case with t>0}
      if permission=1 then delete_columns_with_one_entry;
            {permission=1 iff all smaller blocks previously removed}
      if foundblock=true then delete_columns_with_no_entries;

      if (nrows<s) or (ncols<t) then foundblock:=false;
   
      if foundblock=true then begin
         find_row;    {finds row r0 of Bm with maximum tempval and 
                       adds this row and its intersecting columns to block}
                      {new matrix C now set up with r0 and its columns deleted}

         find_columnblocks(Chead0, Chead1, t-tv, s-1, 0);    {recursive call}

         if foundblock=false then begin   {r0 is not in a (t,s)-block}
             Lhead1[Bnumber[r0]]:='d'; tempval[r0]:=-1;
             block:=block-[r0];          {** all these r0's are OK **}
             ninblock:=ninblock-1;   
             for ii:=1 to nzeros do if (Lhead0[Bnumber[clus0[ii]]]='c') 
             and (clus0[ii] in opmism[r0]) then begin
                 block:=block-[clus0[ii]];
                 ninblock:=ninblock-1;
                        {remove columns intersecting r0 in B from block}
             end;   {ii}     
             find_columnblocks(Lhead0, Lhead1, t, s, 1);
         end;  {if}

      end;    {if}
  
   end;  {else}
  { if foundblock=true then writeln (t:2,'-block');}
end;   {find_columnblocks}

{--------------------------------------------------}



procedure find_rowblocks(Lhead0: header0;  Lhead1: header1; 
                         t, s, permission: integer);   

var nrows, ncols, c0, ii, tv: integer;  
                                  {c0 is a column with maximum valency in Bm}
                 Chead0 : header0;
                 Chead1 : header1;
                tempval : array[1..Ntotal] of integer;
         
{procedure looks for (t,s)-blocks in the ROWS of Bm, i.e. a set of >=s 
columns with entries exclusively within some subset of t rows}

{-----------------------}
  procedure find_tempval;  {finds all valencies for current Bm} 
  var v,w : integer;

  begin  {find_tempval}
{***}{writeln('Tempval....');}
     ncols:=0; nrows:=0; 
     for v:=1 to nzeros do tempval[clus0[v]]:=0;
     for v:=1 to nones do tempval[clus1[v]]:=0;

     for v:=1 to nzeros do begin
       if (Lhead0[Bnumber[clus0[v]]]='c') then begin
          ncols:=ncols+1;
          for w:=1 to nones do
             if (Lhead1[Bnumber[clus1[w]]]='c') 
             and (clus1[w] in opmism[clus0[v]])
             then tempval[clus0[v]]:=tempval[clus0[v]]+1;
       end {header 'c'}
       else tempval[clus0[v]]:=-1;  {header not 'c'}
     end;  {v in 1 to nzeros}
               
     for v:=1 to nones do begin
       if (Lhead1[Bnumber[clus1[v]]]='c') then begin
          nrows:=nrows+1;
          for w:=1 to nzeros do
             if (Lhead0[Bnumber[clus0[w]]]='c') 
             and (clus0[w] in opmism[clus1[v]])
             then tempval[clus1[v]]:=tempval[clus1[v]]+1;
       end {header 'c'}
       else tempval[clus1[v]]:=-1;  {header not 'c'}
     end;  {v in 1 to nones}

    { writeln('found tempval array'); 
    {tempval array found, and ncols, nrows calculated}
 
  end;   {find_tempval}


{-------------------------}
  procedure delete_columns_with_tempval_too_big;

  {columns with tempval>t cannot form part of a (t,s)-block, so are marked 
  'd'}

  var v,w : integer;
 
  begin  {delete_columns_with_tempval_too_big}
{***}{writeln('Columns with tempval too big...');}
  v:=0;
  repeat
     v:=v+1;
     if (Lhead0[Bnumber[clus0[v]]]='c') and (tempval[clus0[v]]>t) then begin
        Lhead0[Bnumber[clus0[v]]]:='d';
        tempval[clus0[v]]:=-1;
        ncols:=ncols-1;
        for w:=1 to nones do if (Lhead1[Bnumber[clus1[w]]]='c') 
        and (clus1[w] in opmism[clus0[v]]) 
        then tempval[clus1[w]]:=tempval[clus1[w]]-1;
     end;  {if v has header 'c' and tempval>t}
  until (v=nzeros) or (ncols<s);
 {writeln('deleted columns with tempval too big');} 
  end;  {delete_columns_with_tempval_too_big}


{-------------------------}
  procedure delete_rows_with_one_entry;
  
 {if permission=1 then we delete all rows with only one entry, and 
 the associated column, as these cannot form part of a block with no smaller 
 subblocks}

  var l : integer;
      someleft : boolean;  

  procedure delete(row : integer);   {row is the MISMATCH number of the site
                                     corresponding to a row with one entry}
    var col, m, done : integer;
    begin               {delete sub-procedure}
       Lhead1[Bnumber[clus1[row]]]:='d'; tempval[clus1[row]]:=-1; 
       nrows:=nrows-1;
       col:=0; done:=0;
       repeat {find col intersecting with row}
          col:=col+1;
          if (Lhead0[Bnumber[clus0[col]]]='c') and 
          (clus0[col] in opmism[clus1[row]]) then done:=1;
       until done=1;
       {delete the intersecting column}
       Lhead0[Bnumber[clus0[col]]]:='d'; tempval[clus0[col]]:=-1; 
       ncols:=ncols-1;

       for m:=1 to nones do if (Lhead1[Bnumber[clus1[m]]]='c') 
       and (clus1[m] in opmism[clus0[col]]) then begin
          tempval[clus1[m]]:=tempval[clus1[m]]-1;
          if (m<row) and (tempval[clus1[m]]=1) then someleft:=true;
                    {m<row iff Bnumber[clus1[m]]<Bnumber[clus1[row]]}
       end;  {m}
    end;  {delete}

    
  begin  {delete_rows_with_one_entry}

{***}{writeln('Rows with one entry...');}
    repeat
       l:=0; someleft:=false;
       repeat
          l:=l+1;
          if (Lhead1[Bnumber[clus1[l]]]='c') and (tempval[clus1[l]]=1) 
          then delete(l);    
          if (nrows<t) or (ncols<s) then foundblock:=false;
       until (l=nones) or (foundblock=false);
    until (someleft=false) or (foundblock=false);
   {writeln('deleted rows with one entry'); }
  end;  {delete_rows_with_one_entry}
  
  
{------------------------}
  procedure delete_rows_with_no_entries; 
  var v : integer;
  begin    {delete_rows_with_no_entries}
{***}{writeln('Rows with no entries.....');}
    for v:=1 to nones do 
       if (Lhead1[Bnumber[clus1[v]]]='c') and (tempval[clus1[v]]=0) 
       then begin
          Lhead1[Bnumber[clus1[v]]]:='d'; tempval[clus1[v]]:=-1; 
          nrows:=nrows-1;
       end;
   {writeln('deleted rows with no entries'); }
  end;     {delete_rows_with_no_entries}


{--------------------}
  procedure find_column;
  {finds row r0 with maximum value of tempval in Bm and adds it together
   with its associated columns to the set block}

   var max,v : integer;
   
   begin  {find_column}
{***}{writeln('Find column....');}
     max:=0;
     for v:=1 to nzeros do if (Lhead0[Bnumber[clus0[v]]]='c') 
     and (tempval[clus0[v]]>max) then begin
         c0:=clus0[v]; max:=tempval[clus0[v]];  
                            {c0 is the MISMATCH number of the column}
     end;  {v}
     {found c0 with maximum tempval amongst columns with header 'c'}
     
     Chead0:=Lhead0; Chead1:=Lhead1;
     Chead0[Bnumber[c0]]:='d';   {delete c0 in new matrix C}
                                       {** these c0's are OK **}
     block:=block+[c0]; ninblock:=ninblock+1;
     tv:=max; tempval[c0]:=-1;
     for v:=1 to nones do if (Lhead1[Bnumber[clus1[v]]]='c') 
     and (clus1[v] in opmism[c0]) then begin
         block:=block+[clus1[v]];
         ninblock:=ninblock+1;
         Chead1[Bnumber[clus1[v]]]:='d'; tempval[clus1[v]]:=-1; 
                                 {delete rows intersecting c0 in C}
     end;   {v}
     {block now contains c0 and its intersected rows}
     {writeln('found column: c0 = ',c0:2); }
   end;    {find_column}

  
{-----------------------------}
  
begin  {main body of find_rowblocks(Lhead0, Lhead1, t, s, permission) }
  { if permission=1 then writeln('looking for ',t:2,'-blocks') 
   else writeln('looking for ',t:2,'-subblocks'); }
   foundblock:=true;
   find_tempval;
   delete_columns_with_tempval_too_big;
                          {now have nrows rows and ncols columns left}

   if t=0 then begin    {degenerate case :t=0}
      if ncols>=s then begin
         foundblock:=true;
         for ii:=1 to nzeros do if (Lhead0[Bnumber[clus0[ii]]]='c') 
         then begin
             block:=block+[clus0[ii]];
             ninblock:=ninblock+1;
         end; {ii}
      end  {ncols>=s}
      else foundblock:=false;
   end   {t=0}
   else begin  {general case with t>0}
      if permission=1 then delete_rows_with_one_entry;
            {permission=1 iff all smaller blocks previously removed}
      if foundblock=true then delete_rows_with_no_entries;

      if (ncols<s) or (nrows<t) then foundblock:=false;
   
      if foundblock=true then begin
         find_column;    {finds column c0 of Bm with maximum tempval and 
                       adds this column and its intersecting rows to block}
                      {new matrix C now set up with c0 and its rows deleted}

         find_rowblocks(Chead0, Chead1, t-tv, s-1, 0);    {recursive call}

         if foundblock=false then begin   {c0 is not in a (t,s)-block}
             Lhead0[Bnumber[c0]]:='d'; tempval[c0]:=-1;
             block:=block-[c0];       {** all these c0's are OK **}
             ninblock:=ninblock-1;   
             for ii:=1 to nones do if (Lhead1[Bnumber[clus1[ii]]]='c')
             and (clus1[ii] in opmism[c0]) then begin
                 block:=block-[clus1[ii]];
                 ninblock:=ninblock-1;
                        {remove rows intersecting c0 in B from block}
             end;   {ii}     
             find_rowblocks(Lhead0, Lhead1, t, s, 1);
         end;  {if}

      end;    {if}
  
   end;  {else}
  { if foundblock=true then writeln (t:2,'-block');}
end;   {find_rowblocks}

{---------------------------------------------------------------}



procedure swap_block(rr:integer);  {for swapping an (rr,rr)-block}
var v,w:integer;

    { *** note that the (t,s)-block here uses different terminology from
     that used in the write-up: there, an (r,s) block had r sites of one
     sort and s>=r of the other; here, a (t,s)-block has t sites of one
     sort and >=s (we don't stipulate how many, just that s is the minimum)
     of the other ***} 


begin  {swap_block}
  { writeln('Block found : elements');}
   {swap block}
   for v:=1 to N do if v in block then begin
        if v in mism0 then head0[Bnumber[v]]:=' ' else head1[Bnumber[v]]:=' ';
        valency[v]:=-1;
        for w:=1 to N do if (w in opmism[v]) and not(w in block) then begin 
            valency[w]:=valency[w]-1;
            opmism[w]:=opmism[w]-[v];
            if valency[w]=1 then singagain:=true;
        end; {w valent to v}
       { writeln(v);}
   end;  {v}
  {all sites in block have been deleted}

   BAM:=BAM-ninblock+2*rr;  {BAM reduced by number unswapped in block}
end;   {swap_block}


{----------------------------------------------------------------------}


{-----THE PROGRAM------------------------------------------------------}

begin

   reset(Nmismfile);
   readln(Nmismfile, Nmism0);
   readln(Nmismfile, Nmism1);
   N:=Nmism0+Nmism1;

   timetostop:=false;
   if Nmism0=0 then timetostop:=true;
   if Nmism1=0 then timetostop:=true;

   if timetostop=true then BAM:=Ntotal-N;
   {read in mism0, mism1, opmism sets, and calculate valencies}
   if timetostop=false then begin
      mism0:=[]; mism1:=[]; 
      for i:=1 to N do opmism[i]:=[];
      reset(mismatch0); reset(mismatch1); reset(opmismatch);
      for i:=1 to Nmism0 do begin
	 read(mismatch0, j);
	 mism0:=mism0+[j];
	 Bnumber[j]:=i;
      end;  {i}
      for i:=1 to Nmism1 do begin
	 read(mismatch1, j); 
	 mism1:=mism1+[j];
	 Bnumber[j]:=i;
      end;  {i}

      for i:=1 to N do begin
	 opmism[i]:=[]; valency[i]:=0;
	 repeat
	    read(opmismatch, j);
	    if j>-1 then begin
	       opmism[i]:=opmism[i]+[j];
	       valency[i]:=valency[i]+1;
	    end;
	 until j=-1;
	 readln(opmismatch);
      end; {i}     

   end;  {if timetostop= false}
   
    {***} writeln('Data read in...'); 

{-----THE BAM ALGORITHM-------------------------------------------------}

  {---set up the headers for B---}

   {mark ALL rows and columns with a 'u' (c.f. the uneconomical B-matrix) 
    since ALL rows and columns of the economical B are used.}
   
   Nsing:=0; nblocks:=0; n2blocks:=0; clusswap:=0;

   if timetostop=false then begin
     BAM:=Ntotal;   
     for i:=1 to Nmism0 do head0[i]:='u';
     for i:=1 to Nmism1 do head1[i]:='u';
      
     singletons;
     Nsing1:=Nsing;  {number of singletons in first round}

      TidyB;
     {writeln('So far... done singletons once and BAM = ',BAM:2);
      if timetostop=true then writeln('timetostop') else writeln('not timetostop');}
  end;  {if timetostop = false: there is now a new value of timetostop}
     
  if timetostop=false then repeat
     
     repeat    {set up connected cliques and check for chains}  
     
        {find lowest k with header 'u'}
         k:=0; timetostop:=true;
         repeat
            k:=k+1;
            if k in mism0 then if head0[Bnumber[k]]='u' then timetostop:=false;
            if k in mism1 then if head1[Bnumber[k]]='u' then timetostop:=false;
         until (timetostop=false) or (k=N);

         if timetostop=false then begin
     {***}{ write('finding connected clique of site ',k:2); }
            find_connected_clique(k);
            check_chains;  {returns a value of nochains: true or false}
     {***}{  writeln('.... & chains, done'); }
          { writeln;
            if nochains=false then writeln('Done a chain')
            else writeln('No more chains'); }
         end;
     until (timetostop=true) or (nochains=true);
  
   {  if timetostop=false then writeln('not timetostop')
     else writeln('timetostop'); }

     if timetostop=false then begin

       {-------the r-block routine----------}
         
       r:=1;
       repeat
         r:=r+1; 
         block:=[]; ninblock:=0; foundblock:=true;
    
         find_columnblocks(head0, head1, r, r, 1);

         singagain:=false;  {swap_block tests to see if necessary to do
                             singletons again}
           
         if foundblock=true then swap_block(r);
      
         if foundblock=false then begin
           { writeln('no ',r:2,'-blocks in columns of B'); }
            block:=[]; ninblock:=0; foundblock:=true; singagain:=false;
            find_rowblocks(head0, head1, r, r, 1);
            if foundblock=true then swap_block(r);
         end;
        {if foundblock=false then writeln('no ',r:2,'-blocks in rows of B');}
       until foundblock=true;   {must terminate eventually}
 
       {----end of the r-block routine-----}
     
        writeln('Swapped ', r:2,'-block'); 
        nblocks:=nblocks+1;
        if r=2 then n2blocks:=n2blocks+1; 
        RestoreB;
        if singagain=true then singletons;
       {writeln('singletons done again, BAM = ',BAM:2);}
        TidyB;  {returns a new value of timetostop}
        
       {if timetostop=true then writeln('timetostop') 
        else writeln('not timetostop'); }

     end;    {r-blocks, singletons and TidyB when timetostop=false}

  until timetostop=true;   

 writeln; writeln;
  writeln('-------------------------------------------------------');
  writeln('SUMMARY:');writeln;
  writeln('Swap attributable to singletons: ',Nsing:2);
  writeln('Swap attributable to perfect clusters: ', clusswap:2);
  writeln('Number of 2-blocks: ',n2blocks:2);
  writeln('Total number of blocks: ',nblocks:2);writeln;
  writeln('Number of sites: ',Ntotal:2);
  writeln('Original match: ',Ntotal-N:2);
  writeln('Best attainable match: ',BAM:2);
  writeln('Best possible swap: ',(BAM-Ntotal+N)/2:1:0);
  writeln; writeln; 
   
end.
