{                BEST ATTAINABLE MATCH ALGORITHM
                          R.M.Fewster
              University of St. Andrews, Scotland, UK.

                           gridgen.p
                           ---------    
Demonstration data-generation program for the Best Attainable Match
algorithm.  This program is free and comes with no guarantees.

Generates random presence/absence maps on a rectangular grid with 
dimensions  XX by YY and orientation 
  
              0 ___ YY
              |
              |
             XX

Clique structure is set up so that the clique of each site is its 8
nearest neighbours, or as many of these as exist.  The clique structure is
written into the file cliquesfile.

Generates 2 maps, the first with Pr(presence)=mix1 for every site and 
the second with Pr(presence)=mix2 for every site.
The maps are written into the file preliminfile.
Edit the program at the points marked *EDIT* to set the dimensions of the 
rectangular grid, XX and YY, and the probabilities of presence for
maps 1 and 2, mix1 and mix2.  Then compile and run.}

program distgenerator(input, preliminfile, cliquesfile);

const XX=50;      {*EDIT*: dimension 1 of rectangular grid} 
      YY=50;      {*EDIT*: dimension 2 of rectangular grid}
      mix1 = 0.5;  {*EDIT*: Pr(presence) for each site in map 1}
      mix2 = 0.5;  {*EDIT*: Pr(presence) for each site in map 2}

var  i,j, x, y, st, wall : integer;
     ran : real;
     preliminfile, cliquesfile: text;

begin
{------generate maps on grid--------}

     rewrite(preliminfile);
     wall:=seed(wallclock);
     for i:=1 to XX do begin
        for j:=1 to YY do begin
           ran:=random(1);
           if ran<=mix1 then write(preliminfile,'1 ') 
           else write(preliminfile,'0 ');
        end;  {j}  {end of jth element of ith row of grid1}
        writeln(preliminfile);
     end;  {i}  {end of ith row of grid1}

     writeln(preliminfile); writeln(preliminfile);
     for i:=1 to XX do begin
        for j:=1 to YY do begin
           ran:=random(1);
           if ran<=mix2 then write(preliminfile,'1 ') 
           else write(preliminfile,'0 ');
        end;  {j}  {end of jth element of ith row of grid2}
        writeln(preliminfile);
     end;  {i}  {end of ith row of grid2}  

{---------set up cliques on grid----------}

     rewrite(cliquesfile);
     writeln(cliquesfile, '  2 ', YY+1:3,' ',YY+2:3,'  -1');
     for y:=2 to YY-1 do 
        writeln(cliquesfile,y-1:3,' ',y+1:3,' ',YY+y-1:3,' ', YY+y:3,' ',
        YY+y+1:3,'  -1');
     writeln(cliquesfile, YY-1:3,' ',2*YY-1:3,' ',2*YY:3,'  -1');
     {end of first edge}
 
     for x:=2 to XX-1 do begin
        st:=(x-1)*YY+1;  {the site number represented by (x,y)}
        writeln(cliquesfile, st-YY:3,' ',st-YY+1:3,' ',st+1:3,' ',
        st+YY:3,' ',st+YY+1:3,'  -1');  {edge site}
        for y:=2 to YY-1 do begin
           st:=(x-1)*YY+y;
           writeln(cliquesfile, st-YY-1:3,' ',st-YY:3,' ',st-YY+1:3,' ',
           st-1:3,' ',st+1:3,' ',st+YY-1:3,' ',st+YY:3,' ',st+YY+1:3,'   -1');
        end; {y}
        st:=x*YY;
        writeln(cliquesfile, st-YY-1:3,' ',st-YY:3,' ',st-1:3,' ',
        st+YY-1:3,' ',st+YY:3,'  -1');  {edge site}
     end; {x}

     st:=(XX-1)*YY+1;
     writeln(cliquesfile,  st-YY:3,' ',st-YY+1:3,' ',st+1:3,'  -1');
     for y:=2 to YY-1 do begin
        st:=(XX-1)*YY+y;
        writeln(cliquesfile, st-YY-1:3,' ',st-YY:3,' ',st-YY+1:3,' ',
        st-1:3,' ',st+1:3,'   -1');
     end;  {y}
     writeln(cliquesfile, (XX-1)*YY-1:3,' ',(XX-1)*YY:3,' ',XX*YY-1:3,'  -1');
     {end of last edge}     


end.


